# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "Provider",
    "State",
    "LibcloudLBError",
    "LibcloudLBImmutableError",
    "OLD_CONSTANT_TO_NEW_MAPPING",
]

from libcloud.common.types import LibcloudError


class LibcloudLBError(LibcloudError):
    pass


class LibcloudLBImmutableError(LibcloudLBError):
    pass


class Provider:
    """
    Defines for each of the supported providers

    Non-Dummy drivers are sorted in alphabetical order. Please preserve this
    ordering when adding new drivers.

    :cvar ALIYUN_SLB: Aliyun SLB loadbalancer driver
    """

    ALB = "alb"
    ALIYUN_SLB = "aliyun_slb"
    BRIGHTBOX = "brightbox"
    CLOUDSTACK = "cloudstack"
    DIMENSIONDATA = "dimensiondata"
    ELB = "elb"
    GCE = "gce"
    NINEFOLD = "ninefold"
    NTTCIS = "nttcis"
    RACKSPACE = "rackspace"

    # Deprecated
    RACKSPACE_US = "rackspace_us"
    RACKSPACE_UK = "rackspace_uk"


OLD_CONSTANT_TO_NEW_MAPPING = {
    Provider.RACKSPACE_US: Provider.RACKSPACE,
    Provider.RACKSPACE_UK: Provider.RACKSPACE,
}


class State:
    """
    Standard states for a loadbalancer

    :cvar RUNNING: loadbalancer is running and ready to use
    :cvar UNKNOWN: loabalancer state is unknown
    """

    RUNNING = 0
    PENDING = 1
    UNKNOWN = 2
    ERROR = 3
    DELETED = 4


class MemberCondition:
    """
    Each member of a load balancer can have an associated condition
    which determines its role within the load balancer.
    """

    ENABLED = 0
    DISABLED = 1
    DRAINING = 2



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys

import pytest

from libcloud.test import MockHttp, unittest
from libcloud.utils.py3 import httplib
from libcloud.common.types import InvalidCredsError
from libcloud.test.secrets import NTTCIS_PARAMS
from libcloud.common.nttcis import NttCisPool, NttCisVIPNode, NttCisPoolMember, NttCisAPIException
from libcloud.loadbalancer.base import Member, Algorithm, LoadBalancer
from libcloud.loadbalancer.types import State
from libcloud.test.file_fixtures import LoadBalancerFileFixtures
from libcloud.loadbalancer.drivers.nttcis import NttCisLBDriver


@pytest.fixture()
def driver():
    NttCisLBDriver.connectionCls.active_api_version = "2.7"
    NttCisLBDriver.connectionCls.conn_class = NttCisMockHttp
    NttCisMockHttp.type = None
    driver = NttCisLBDriver(*NTTCIS_PARAMS)
    return driver


def test_invalid_region(driver):
    with pytest.raises(ValueError):
        NttCisLBDriver(*NTTCIS_PARAMS, region="blah")


def test_invalid_creds(driver):
    NttCisMockHttp.type = "UNAUTHORIZED"
    with pytest.raises(InvalidCredsError):
        driver.list_balancers()


def test_create_balancer(driver, capsys):
    print(Algorithm.ROUND_ROBIN)
    driver.ex_set_current_network_domain("1234")
    members = []
    members.append(Member(id=None, ip="1.2.3.4", port=80))

    balancer = driver.create_balancer(
        name="test",
        port=80,
        protocol="http",
        optimization_profile="TCP",
        algorithm=Algorithm.ROUND_ROBIN,
        members=members,
        ex_listener_ip_address="5.6.7.8",
    )
    assert balancer.name == "test"
    assert balancer.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert balancer.ip == "165.180.12.22"
    assert balancer.port == 80
    assert balancer.extra["pool_id"] == "9e6b496d-5261-4542-91aa-b50c7f569c54"
    assert balancer.extra["network_domain_id"] == "1234"
    assert balancer.extra["listener_ip_address"] == "5.6.7.8"


def test_create_balancer_with_defaults(driver):
    driver.ex_set_current_network_domain("1234")

    balancer = driver.create_balancer(
        name="test", port=None, protocol=None, algorithm=None, members=None
    )
    assert balancer.name == "test"
    assert balancer.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert balancer.ip == "165.180.12.22"
    assert balancer.port is None
    assert balancer.extra["pool_id"] == "9e6b496d-5261-4542-91aa-b50c7f569c54"
    assert balancer.extra["network_domain_id"] == "1234"


def test_create_balancer_no_members(driver):
    driver.ex_set_current_network_domain("1234")
    members = None

    balancer = driver.create_balancer(
        name="test",
        port=80,
        protocol="http",
        algorithm=Algorithm.ROUND_ROBIN,
        members=members,
    )
    assert balancer.name == "test"
    assert balancer.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert balancer.ip == "165.180.12.22"
    assert balancer.port == 80
    assert balancer.extra["pool_id"] == "9e6b496d-5261-4542-91aa-b50c7f569c54"
    assert balancer.extra["network_domain_id"] == "1234"


def test_create_balancer_empty_members(driver):
    driver.ex_set_current_network_domain("1234")
    members = []

    balancer = driver.create_balancer(
        name="test",
        port=80,
        protocol="http",
        algorithm=Algorithm.ROUND_ROBIN,
        members=members,
    )
    assert balancer.name == "test"
    assert balancer.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert balancer.ip == "165.180.12.22"
    assert balancer.port == 80
    assert balancer.extra["pool_id"] == "9e6b496d-5261-4542-91aa-b50c7f569c54"
    assert balancer.extra["network_domain_id"] == "1234"


def test_list_balancers(driver):
    bal = driver.list_balancers()
    assert bal[0].name == "myProduction.Virtual.Listener"
    assert bal[0].id == "6115469d-a8bb-445b-bb23-d23b5283f2b9"
    assert bal[0].port == "8899"
    assert bal[0].ip == "165.180.12.22"
    assert bal[0].state == State.RUNNING


def test_balancer_list_members(driver):
    extra = {
        "pool_id": "4d360b1f-bc2c-4ab7-9884-1f03ba2768f7",
        "network_domain_id": "1234",
    }
    balancer = LoadBalancer(
        id="234",
        name="test",
        state=State.RUNNING,
        ip="1.2.3.4",
        port=1234,
        driver=driver,
        extra=extra,
    )
    members = driver.balancer_list_members(balancer)
    assert 2 == len(members)
    assert members[0].ip == "10.0.3.13"
    assert members[0].id == "3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0"
    assert members[0].port == 9889


def test_balancer_attach_member(driver):
    extra = {
        "pool_id": "4d360b1f-bc2c-4ab7-9884-1f03ba2768f7",
        "network_domain_id": "1234",
    }
    balancer = LoadBalancer(
        id="234",
        name="test",
        state=State.RUNNING,
        ip="1.2.3.4",
        port=1234,
        driver=driver,
        extra=extra,
    )
    member = Member(id=None, ip="112.12.2.2", port=80, balancer=balancer, extra=None)
    member = driver.balancer_attach_member(balancer, member)
    assert member.id == "3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0"


def test_balancer_attach_member_without_port(driver):
    extra = {
        "pool_id": "4d360b1f-bc2c-4ab7-9884-1f03ba2768f7",
        "network_domain_id": "1234",
    }
    balancer = LoadBalancer(
        id="234",
        name="test",
        state=State.RUNNING,
        ip="1.2.3.4",
        port=1234,
        driver=driver,
        extra=extra,
    )
    member = Member(id=None, ip="112.12.2.2", port=None, balancer=balancer, extra=None)
    member = driver.balancer_attach_member(balancer, member)
    assert member.id == "3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0"
    assert member.port is None


def test_balancer_detach_member(driver):
    extra = {
        "pool_id": "4d360b1f-bc2c-4ab7-9884-1f03ba2768f7",
        "network_domain_id": "1234",
    }
    balancer = LoadBalancer(
        id="234",
        name="test",
        state=State.RUNNING,
        ip="1.2.3.4",
        port=1234,
        driver=driver,
        extra=extra,
    )
    member = Member(
        id="3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0",
        ip="112.12.2.2",
        port=80,
        balancer=balancer,
        extra=None,
    )
    result = driver.balancer_detach_member(balancer, member)
    assert result, True


def test_destroy_balancer(driver):
    extra = {
        "pool_id": "4d360b1f-bc2c-4ab7-9884-1f03ba2768f7",
        "network_domain_id": "1234",
    }
    balancer = LoadBalancer(
        id="234",
        name="test",
        state=State.RUNNING,
        ip="1.2.3.4",
        port=1234,
        driver=driver,
        extra=extra,
    )
    response = driver.destroy_balancer(balancer)
    assert response is True


def test_set_get_network_domain_id(driver):
    driver.ex_set_current_network_domain("1234")
    nwd = driver.ex_get_current_network_domain()
    assert nwd == "1234"


def test_ex_create_pool_member(driver):
    pool = NttCisPool(
        id="4d360b1f-bc2c-4ab7-9884-1f03ba2768f7",
        name="test",
        description="test",
        status=State.RUNNING,
        health_monitor_id=None,
        load_balance_method=None,
        service_down_action=None,
        slow_ramp_time=None,
    )
    node = NttCisVIPNode(id="2344", name="test", status=State.RUNNING, ip="123.23.3.2")
    member = driver.ex_create_pool_member(pool=pool, node=node, port=80)
    assert member.id == "3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0"
    assert member.name == "10.0.3.13"
    assert member.ip == "123.23.3.2"


def test_ex_create_node(driver):
    node = driver.ex_create_node(
        network_domain_id="12345",
        name="test",
        ip="123.12.32.2",
        ex_description="",
        connection_limit=25000,
        connection_rate_limit=2000,
    )
    assert node.name == "myProductionNode.1"
    assert node.id == "9e6b496d-5261-4542-91aa-b50c7f569c54"


def test_ex_create_pool(
    driver,
):
    pool = driver.ex_create_pool(
        network_domain_id="1234",
        name="test",
        balancer_method="ROUND_ROBIN",
        ex_description="test",
        service_down_action="NONE",
        slow_ramp_time=30,
    )
    assert pool.id == "9e6b496d-5261-4542-91aa-b50c7f569c54"
    assert pool.name == "test"
    assert pool.status == State.RUNNING


def test_ex_create_virtual_listener(driver):
    listener = driver.ex_create_virtual_listener(
        network_domain_id="12345",
        name="test",
        ex_description="test",
        port=80,
        pool=NttCisPool(
            id="1234",
            name="test",
            description="test",
            status=State.RUNNING,
            health_monitor_id=None,
            load_balance_method=None,
            service_down_action=None,
            slow_ramp_time=None,
        ),
    )
    assert listener.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert listener.name == "test"


def test_ex_create_virtual_listener_unusual_port(driver):
    listener = driver.ex_create_virtual_listener(
        network_domain_id="12345",
        name="test",
        ex_description="test",
        port=8900,
        pool=NttCisPool(
            id="1234",
            name="test",
            description="test",
            status=State.RUNNING,
            health_monitor_id=None,
            load_balance_method=None,
            service_down_action=None,
            slow_ramp_time=None,
        ),
    )
    assert listener.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert listener.name == "test"


def test_ex_create_virtual_listener_without_port(driver):
    listener = driver.ex_create_virtual_listener(
        network_domain_id="12345",
        name="test",
        ex_description="test",
        pool=NttCisPool(
            id="1234",
            name="test",
            description="test",
            status=State.RUNNING,
            health_monitor_id=None,
            load_balance_method=None,
            service_down_action=None,
            slow_ramp_time=None,
        ),
    )
    assert listener.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert listener.name == "test"


def test_ex_create_virtual_listener_without_pool(driver):
    listener = driver.ex_create_virtual_listener(
        network_domain_id="12345", name="test", ex_description="test"
    )
    assert listener.id == "8334f461-0df0-42d5-97eb-f4678eb26bea"
    assert listener.name == "test"


def test_get_balancer(driver):
    bal = driver.get_balancer("6115469d-a8bb-445b-bb23-d23b5283f2b9")
    assert bal.name == "myProduction.Virtual.Listener"
    assert bal.id == "6115469d-a8bb-445b-bb23-d23b5283f2b9"
    assert bal.port == "8899"
    assert bal.ip == "165.180.12.22"
    assert bal.state == State.RUNNING


def test_list_protocols(driver):
    protocols = driver.list_protocols()
    assert 0 < len(protocols)


def test_ex_get_nodes(driver):
    nodes = driver.ex_get_nodes()
    assert 2 == len(nodes)
    assert nodes[0].name == "ProductionNode.1"
    assert nodes[0].id == "34de6ed6-46a4-4dae-a753-2f8d3840c6f9"
    assert nodes[0].ip == "10.10.10.101"


def test_ex_get_node(driver):
    node = driver.ex_get_node("34de6ed6-46a4-4dae-a753-2f8d3840c6f9")
    assert node.name == "ProductionNode.2"
    assert node.id == "34de6ed6-46a4-4dae-a753-2f8d3840c6f9"
    assert node.ip == "10.10.10.101"


def test_ex_update_node(driver):
    node = driver.ex_get_node("34de6ed6-46a4-4dae-a753-2f8d3840c6f9")
    node.connection_limit = "100"
    result = driver.ex_update_node(node)
    assert result.connection_limit == "100"


def test_ex_destroy_node(driver):
    result = driver.ex_destroy_node("34de6ed6-46a4-4dae-a753-2f8d3840c6f9")
    assert result is True


def test_ex_set_node_state(driver):
    node = driver.ex_get_node("34de6ed6-46a4-4dae-a753-2f8d3840c6f9")
    result = driver.ex_set_node_state(node, False)
    assert result.connection_limit == "10000"


def test_ex_get_pools(driver):
    pools = driver.ex_get_pools()
    assert 0 != len(pools)
    assert pools[0].name == "myDevelopmentPool.1"
    assert pools[0].id == "4d360b1f-bc2c-4ab7-9884-1f03ba2768f7"


def test_ex_get_pool(driver):
    pool = driver.ex_get_pool("4d360b1f-bc2c-4ab7-9884-1f03ba2768f7")
    assert pool.name == "myDevelopmentPool.1"
    assert pool.id == "4d360b1f-bc2c-4ab7-9884-1f03ba2768f7"


def test_ex_update_pool(driver):
    pool = driver.ex_get_pool("4d360b1f-bc2c-4ab7-9884-1f03ba2768f7")
    pool.slow_ramp_time = "120"
    result = driver.ex_update_pool(pool)
    assert result is True


def test_ex_destroy_pool(driver):
    response = driver.ex_destroy_pool(
        pool=NttCisPool(
            id="4d360b1f-bc2c-4ab7-9884-1f03ba2768f7",
            name="test",
            description="test",
            status=State.RUNNING,
            health_monitor_id=None,
            load_balance_method=None,
            service_down_action=None,
            slow_ramp_time=None,
        )
    )
    assert response is True


def test_get_pool_members(driver):
    members = driver.ex_get_pool_members("4d360b1f-bc2c-4ab7-9884-1f03ba2768f7")
    assert 2 == len(members)
    assert members[0].id == "3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0"
    assert members[0].name == "10.0.3.13"
    assert members[0].status == "NORMAL"
    assert members[0].ip == "10.0.3.13"
    assert members[0].port == 9889
    assert members[0].node_id == "3c207269-e75e-11e4-811f-005056806999"


def test_get_pool_member(driver):
    member = driver.ex_get_pool_member("3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0")
    assert member.id == "3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0"
    assert member.name == "10.0.3.13"
    assert member.status == "NORMAL"
    assert member.ip == "10.0.3.13"
    assert member.port == 9889


def test_set_pool_member_state(driver):
    member = driver.ex_get_pool_member("3dd806a2-c2c8-4c0c-9a4f-5219ea9266c0")
    result = driver.ex_set_pool_member_state(member, True)
    assert result is True


def test_ex_destroy_pool_member(driver):
    response = driver.ex_destroy_pool_member(
        member=NttCisPoolMember(
            id="",
            name="test",
            status=State.RUNNING,
            ip="1.2.3.4",
            port=80,
            node_id="3c207269-e75e-11e4-811f-005056806999",
        ),
        destroy_node=False,
    )
    assert response is True


def test_ex_destroy_pool_member_with_node(driver):
    response = driver.ex_destroy_pool_member(
        member=NttCisPoolMember(
            id="",
            name="test",
            status=State.RUNNING,
            ip="1.2.3.4",
            port=80,
            node_id="34de6ed6-46a4-4dae-a753-2f8d3840c6f9",
        ),
        destroy_node=True,
    )
    assert response is True


def test_ex_get_default_health_monitors(driver):
    monitors = driver.ex_get_default_health_monitors("4d360b1f-bc2c-4ab7-9884-1f03ba2768f7")
    assert len(monitors) == 6
    assert monitors[0].id == "01683574-d487-11e4-811f-005056806999"
    assert monitors[0].name == "CCDEFAULT.Http"
    assert monitors[0].node_compatible is False
    assert monitors[0].pool_compatible is True


def test_ex_get_default_persistence_profiles(driver):
    profiles = driver.ex_get_default_persistence_profiles("4d360b1f-bc2c-4ab7-9884-1f03ba2768f7")
    assert len(profiles) == 4
    assert profiles[0].id == "a34ca024-f3db-11e4-b010-005056806999"
    assert profiles[0].name == "CCDEFAULT.Cookie"
    assert profiles[0].fallback_compatible is False
    assert len(profiles[0].compatible_listeners) == 1
    assert profiles[0].compatible_listeners[0].type == "PERFORMANCE_LAYER_4"


def test_ex_get_default_irules(driver):
    irules = driver.ex_get_default_irules("4d360b1f-bc2c-4ab7-9884-1f03ba2768f7")
    assert len(irules) == 4
    assert irules[0].id == "2b20cb2c-ffdc-11e4-b010-005056806999"
    assert irules[0].name == "CCDEFAULT.HttpsRedirect"
    assert len(irules[0].compatible_listeners) == 1
    assert irules[0].compatible_listeners[0].type == "PERFORMANCE_LAYER_4"


def test_ex_insert_ssl_certificate(driver):
    net_dom_id = "6aafcf08-cb0b-432c-9c64-7371265db086 "
    cert = (
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        + "/loadbalancer/fixtures/nttcis/alice.crt"
    )
    key = (
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        + "/loadbalancer/fixtures/nttcis/alice.key"
    )
    result = driver.ex_import_ssl_domain_certificate(
        net_dom_id, "alice", cert, key, description="test cert"
    )
    assert result is True


def test_ex_insert_ssl_certificate_FAIL(driver):
    NttCisMockHttp.type = "FAIL"
    net_dom_id = "6aafcf08-cb0b-432c-9c64-7371265db086 "
    cert = (
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        + "/loadbalancer/fixtures/nttcis/denis.crt"
    )
    key = (
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        + "/loadbalancer/fixtures/nttcis/denis.key"
    )
    with pytest.raises(NttCisAPIException) as excinfo:
        driver.ex_import_ssl_domain_certificate(
            net_dom_id, "denis", cert, key, description="test cert"
        )
    assert (
        excinfo.value.msg == "Data Center EU6 requires key length must be one of 512, 1024, 2048."
    )


def test_ex_create_ssl_offload_profile(driver):
    net_domain_id = "6aafcf08-cb0b-432c-9c64-7371265db086"
    name = "ssl_offload"
    domain_cert = driver.ex_list_ssl_domain_certs(name="alice")[0]
    result = driver.ex_create_ssl_offload_profile(
        net_domain_id, name, domain_cert.id, ciphers="!ECDHE+AES-GCM:"
    )
    assert result is True


def test_ex_list_ssl_offload_profile(driver):
    NttCisMockHttp.type = "LIST"
    profiles = driver.ex_list_ssl_offload_profiles()
    assert profiles[0].sslDomainCertificate.name == "alice"


def test_ex_get_ssl_offload_profile(driver):
    profile_id = "b1d3b5a7-75d7-4c44-a2b7-5bfa773dec63"
    profile = driver.ex_get_ssl_offload_profile(profile_id)
    assert profile.name == "ssl_offload"


def test_edit_ssl_offload_profile(driver):
    profile_name = "ssl_offload"
    datacenter_id = "EU6"
    NttCisMockHttp.type = "LIST"
    profile = driver.ex_list_ssl_offload_profiles(name=profile_name, datacenter_id=datacenter_id)[0]
    NttCisMockHttp.type = None
    result = driver.ex_edit_ssl_offload_profile(
        profile.id,
        profile.name,
        profile.sslDomainCertificate.id,
        ciphers=profile.ciphers,
        description="A test edit of an offload profile",
    )
    assert result is True


def test_delete_ssl_offload_profile(driver):
    profile_name = "ssl_offload"
    NttCisMockHttp.type = "LIST"
    profile = driver.ex_list_ssl_offload_profiles(name=profile_name)[0]
    NttCisMockHttp.type = None
    result = driver.ex_delete_ssl_offload_profile(profile.id)
    assert result is True


def test_delete_ssl_certificate_chain(driver):
    NttCisMockHttp.type = "LIST"
    chain_name = "ted_carol"
    cert_chain = driver.ex_list_ssl_certificate_chains(name=chain_name)[0]
    NttCisMockHttp.type = None
    result = driver.ex_delete_ssl_certificate_chain(cert_chain.id)
    assert result is True


def test_delete_ssl_domain_certificate(driver):
    NttCisMockHttp.type = "LIST"
    cert_name = "alice"
    cert = driver.ex_list_ssl_domain_certs(name=cert_name)[0]
    NttCisMockHttp.type = None
    result = driver.ex_delete_ssl_domain_certificate(cert.id)
    assert result is True


class NttCisMockHttp(MockHttp):
    fixtures = LoadBalancerFileFixtures("nttcis")

    def _oec_0_9_myaccount_UNAUTHORIZED(self, method, url, body, headers):
        return (httplib.UNAUTHORIZED, "", {}, httplib.responses[httplib.UNAUTHORIZED])

    def _oec_0_9_myaccount(self, method, url, body, headers):
        body = self.fixtures.load("oec_0_9_myaccount.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_FAIL(self, method, url, body, headers):
        body = self.fixtures.load("oec_0_9_myaccount.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _oec_0_9_myaccount_LIST(self, method, url, body, headers):
        body = self.fixtures.load("oec_0_9_myaccount.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_virtualListener(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_virtualListener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_virtualListener_6115469d_a8bb_445b_bb23_d23b5283f2b9(
        self, method, url, body, headers
    ):
        body = self.fixtures.load(
            "networkDomainVip_virtualListener_6115469d_a8bb_445b_bb23_d23b5283f2b9.xml"
        )
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_pool(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_pool.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_pool_4d360b1f_bc2c_4ab7_9884_1f03ba2768f7(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_pool_4d360b1f_bc2c_4ab7_9884_1f03ba2768f7.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_poolMember(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_poolMember.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_poolMember_3dd806a2_c2c8_4c0c_9a4f_5219ea9266c0(
        self, method, url, body, headers
    ):
        body = self.fixtures.load(
            "networkDomainVip_poolMember_3dd806a2_c2c8_4c0c_9a4f_5219ea9266c0.xml"
        )
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_createPool(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_createPool.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_createNode(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_createNode.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_addPoolMember(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_addPoolMember.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_createVirtualListener(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_createVirtualListener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_removePoolMember(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_removePoolMember.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_deleteVirtualListener(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_deleteVirtualListener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_deletePool(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_deletePool.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_deleteNode(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_deleteNode.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_node(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_node.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_node_34de6ed6_46a4_4dae_a753_2f8d3840c6f9(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_node_34de6ed6_46a4_4dae_a753_2f8d3840c6f9.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_editNode(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_editNode.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_editPool(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_editPool.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_editPoolMember(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_editPoolMember.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_defaultHealthMonitor(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_defaultHealthMonitor.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_defaultPersistenceProfile(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_defaultPersistenceProfile.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_defaultIrule(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("networkDomainVip_defaultIrule.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_importSslDomainCertificate(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("ssl_import_success.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_importSslDomainCertificate_FAIL(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("ssl_import_fail.xml")
        return (httplib.BAD_REQUEST, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_sslDomainCertificate_LIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("ssl_cert_by_name.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_sslCertificateChain_LIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("ssl_list_cert_chain_by_name.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_sslDomainCertificate(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("ssl_cert_by_name.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_createSslOffloadProfile(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("create_ssl_offload_profile.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_sslOffloadProfile_LIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("list_ssl_offload_profiles.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_sslOffloadProfile_b1d3b5a7_75d7_4c44_a2b7_5bfa773dec63(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("get_ssl_offload_profile.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_editSslOffloadProfile(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("edit_ssl_offload_profile.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_deleteSslOffloadProfile(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("delete_ssl_offload_profile.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_deleteSslCertificateChain(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("delete_ssl_certificate_chain.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _caas_2_7_8a8f6abc_2745_4d8a_9cbc_8dabe5a7d0e4_networkDomainVip_deleteSslDomainCertificate(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("delete_ssl_domain_certificate.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import LB_BRIGHTBOX_PARAMS
from libcloud.loadbalancer.base import Member, Algorithm
from libcloud.loadbalancer.types import State
from libcloud.test.file_fixtures import LoadBalancerFileFixtures
from libcloud.loadbalancer.drivers.brightbox import BrightboxLBDriver


class BrightboxLBTests(unittest.TestCase):
    def setUp(self):
        BrightboxLBDriver.connectionCls.conn_class = BrightboxLBMockHttp
        BrightboxLBMockHttp.type = None
        BrightboxLBDriver.connectionCls.token = "test"
        self.driver = BrightboxLBDriver(*LB_BRIGHTBOX_PARAMS)

    def test_list_protocols(self):
        protocols = self.driver.list_protocols()

        self.assertEqual(len(protocols), 2)
        self.assertTrue("tcp" in protocols)
        self.assertTrue("http" in protocols)

    def test_list_balancers(self):
        balancers = self.driver.list_balancers()

        self.assertEqual(len(balancers), 1)
        self.assertEqual(balancers[0].id, "lba-1235f")
        self.assertEqual(balancers[0].name, "lb1")

    def test_get_balancer(self):
        balancer = self.driver.get_balancer(balancer_id="lba-1235f")

        self.assertEqual(balancer.id, "lba-1235f")
        self.assertEqual(balancer.name, "lb1")
        self.assertEqual(balancer.state, State.RUNNING)

    def test_destroy_balancer(self):
        balancer = self.driver.get_balancer(balancer_id="lba-1235f")

        self.assertTrue(self.driver.destroy_balancer(balancer))

    def test_create_balancer(self):
        members = [Member("srv-lv426", None, None)]

        balancer = self.driver.create_balancer(
            name="lb2",
            port=80,
            protocol="http",
            algorithm=Algorithm.ROUND_ROBIN,
            members=members,
        )

        self.assertEqual(balancer.name, "lb2")
        self.assertEqual(balancer.port, 80)
        self.assertEqual(balancer.state, State.PENDING)

    def test_balancer_list_members(self):
        balancer = self.driver.get_balancer(balancer_id="lba-1235f")
        members = balancer.list_members()

        self.assertEqual(len(members), 1)
        self.assertEqual(members[0].balancer, balancer)
        self.assertEqual("srv-lv426", members[0].id)

    def test_balancer_attach_member(self):
        balancer = self.driver.get_balancer(balancer_id="lba-1235f")
        member = balancer.attach_member(Member("srv-kg983", ip=None, port=None))

        self.assertEqual(member.id, "srv-kg983")

    def test_balancer_detach_member(self):
        balancer = self.driver.get_balancer(balancer_id="lba-1235f")
        member = Member("srv-lv426", None, None)

        self.assertTrue(balancer.detach_member(member))


class BrightboxLBMockHttp(MockHttp):
    fixtures = LoadBalancerFileFixtures("brightbox")

    def _token(self, method, url, body, headers):
        if method == "POST":
            return (
                httplib.OK,
                self.fixtures.load("token.json"),
                {"content-type": "application/json"},
                httplib.responses[httplib.OK],
            )

    def _1_0_load_balancers(self, method, url, body, headers):
        if method == "GET":
            return (
                httplib.OK,
                self.fixtures.load("load_balancers.json"),
                {"content-type": "application/json"},
                httplib.responses[httplib.OK],
            )
        elif method == "POST":
            body = self.fixtures.load("load_balancers_post.json")
            return (
                httplib.ACCEPTED,
                body,
                {"content-type": "application/json"},
                httplib.responses[httplib.ACCEPTED],
            )

    def _1_0_load_balancers_lba_1235f(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("load_balancers_lba_1235f.json")
            return (
                httplib.OK,
                body,
                {"content-type": "application/json"},
                httplib.responses[httplib.OK],
            )
        elif method == "DELETE":
            return (
                httplib.ACCEPTED,
                "",
                {"content-type": "application/json"},
                httplib.responses[httplib.ACCEPTED],
            )

    def _1_0_load_balancers_lba_1235f_add_nodes(self, method, url, body, headers):
        if method == "POST":
            return (
                httplib.ACCEPTED,
                "",
                {"content-type": "application/json"},
                httplib.responses[httplib.ACCEPTED],
            )

    def _1_0_load_balancers_lba_1235f_remove_nodes(self, method, url, body, headers):
        if method == "POST":
            return (
                httplib.ACCEPTED,
                "",
                {"content-type": "application/json"},
                httplib.responses[httplib.ACCEPTED],
            )


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.types import RecordType, ZoneDoesNotExistError, RecordDoesNotExistError
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import DNS_PARAMS_LINODE, DNS_KEYWORD_PARAMS_LINODE
from libcloud.common.linode import LinodeException
from libcloud.dns.drivers.linode import LinodeDNSDriver, LinodeDNSDriverV3
from libcloud.test.file_fixtures import DNSFileFixtures


class LinodeTests(unittest.TestCase):
    def setUp(self):
        LinodeDNSDriverV3.connectionCls.conn_class = LinodeMockHttp
        LinodeMockHttp.use_param = "api_action"
        LinodeMockHttp.type = None
        self.driver = LinodeDNSDriver(*DNS_PARAMS_LINODE, **DNS_KEYWORD_PARAMS_LINODE)

    def assertHasKeys(self, dictionary, keys):
        for key in keys:
            self.assertTrue(key in dictionary, 'key "%s" not in dictionary' % (key))

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 7)
        self.assertTrue(RecordType.A in record_types)

    def test_list_zones_success(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)

        zone = zones[0]
        self.assertEqual(zone.id, "5093")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "linode.com")
        self.assertIsNone(zone.ttl)
        self.assertHasKeys(zone.extra, ["description", "SOA_Email", "status"])

    def test_list_records_success(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 2)

        arecord = records[0]
        self.assertEqual(arecord.id, "3585100")
        self.assertEqual(arecord.name, "mc")
        self.assertEqual(arecord.type, RecordType.A)
        self.assertEqual(arecord.data, "127.0.0.1")
        self.assertHasKeys(arecord.extra, ["protocol", "ttl_sec", "port", "weight"])

        srvrecord = records[1]
        self.assertEqual(srvrecord.id, "3585141")
        self.assertEqual(srvrecord.name, "_minecraft._udp")
        self.assertEqual(srvrecord.type, RecordType.SRV)
        self.assertEqual(srvrecord.data, "mc.linode.com")
        self.assertHasKeys(srvrecord.extra, ["protocol", "ttl_sec", "port", "priority", "weight"])

    def test_list_records_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        LinodeMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.list_records(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_get_zone_success(self):
        LinodeMockHttp.type = "GET_ZONE"

        zone = self.driver.get_zone(zone_id="5093")
        self.assertEqual(zone.id, "5093")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.domain, "linode.com")
        self.assertIsNone(zone.ttl)
        self.assertHasKeys(zone.extra, ["description", "SOA_Email", "status"])

    def test_get_zone_does_not_exist(self):
        LinodeMockHttp.type = "GET_ZONE_DOES_NOT_EXIST"

        try:
            self.driver.get_zone(zone_id="4444")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "4444")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_success(self):
        LinodeMockHttp.type = "GET_RECORD"
        record = self.driver.get_record(zone_id="1234", record_id="3585100")
        self.assertEqual(record.id, "3585100")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.0.1")
        self.assertHasKeys(record.extra, ["protocol", "ttl_sec", "port", "weight"])

    def test_get_record_zone_does_not_exist(self):
        LinodeMockHttp.type = "GET_RECORD_ZONE_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="444", record_id="3585100")
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_record_record_does_not_exist(self):
        LinodeMockHttp.type = "GET_RECORD_RECORD_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="4441", record_id="3585100")
        except RecordDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_zone_success(self):
        zone = self.driver.create_zone(domain="foo.bar.com", type="master", ttl=None, extra=None)
        self.assertEqual(zone.id, "5094")
        self.assertEqual(zone.domain, "foo.bar.com")

    def test_create_zone_validaton_error(self):
        LinodeMockHttp.type = "VALIDATION_ERROR"

        try:
            self.driver.create_zone(domain="foo.bar.com", type="master", ttl=None, extra=None)
        except LinodeException:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_update_zone_success(self):
        zone = self.driver.list_zones()[0]
        updated_zone = self.driver.update_zone(
            zone=zone,
            domain="libcloud.org",
            ttl=10,
            extra={"SOA_Email": "bar@libcloud.org"},
        )

        self.assertEqual(zone.extra["SOA_Email"], "dns@example.com")

        self.assertEqual(updated_zone.id, zone.id)
        self.assertEqual(updated_zone.domain, "libcloud.org")
        self.assertEqual(updated_zone.type, zone.type)
        self.assertEqual(updated_zone.ttl, 10)
        self.assertEqual(updated_zone.extra["SOA_Email"], "bar@libcloud.org")
        self.assertEqual(updated_zone.extra["status"], zone.extra["status"])
        self.assertEqual(updated_zone.extra["description"], zone.extra["description"])

    def test_create_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record(
            name="www", zone=zone, type=RecordType.A, data="127.0.0.1"
        )

        self.assertEqual(record.id, "3585100")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.0.0.1")

    def test_update_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        updated_record = self.driver.update_record(
            record=record, name="www", type=RecordType.AAAA, data="::1"
        )

        self.assertEqual(record.data, "127.0.0.1")

        self.assertEqual(updated_record.id, record.id)
        self.assertEqual(updated_record.name, "www")
        self.assertEqual(updated_record.zone, record.zone)
        self.assertEqual(updated_record.type, RecordType.AAAA)
        self.assertEqual(updated_record.data, "::1")

    def test_delete_zone_success(self):
        zone = self.driver.list_zones()[0]
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        LinodeMockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_does_not_exist(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]

        LinodeMockHttp.type = "RECORD_DOES_NOT_EXIST"

        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, record.id)
        else:
            self.fail("Exception was not thrown")


class LinodeMockHttp(MockHttp):
    fixtures = DNSFileFixtures("linode")

    def _domain_list(self, method, url, body, headers):
        body = self.fixtures.load("domain_list.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _domain_resource_list(self, method, url, body, headers):
        body = self.fixtures.load("resource_list.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _ZONE_DOES_NOT_EXIST_domain_resource_list(self, method, url, body, headers):
        body = self.fixtures.load("resource_list_does_not_exist.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_ZONE_domain_list(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_ZONE_DOES_NOT_EXIST_domain_list(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_does_not_exist.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_RECORD_domain_list(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_RECORD_domain_resource_list(self, method, url, body, headers):
        body = self.fixtures.load("get_record.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_RECORD_ZONE_DOES_NOT_EXIST_domain_list(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_does_not_exist.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_RECORD_ZONE_DOES_NOT_EXIST_domain_resource_list(self, method, url, body, headers):
        body = self.fixtures.load("get_record_does_not_exist.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_RECORD_RECORD_DOES_NOT_EXIST_domain_list(self, method, url, body, headers):
        body = self.fixtures.load("get_zone.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _GET_RECORD_RECORD_DOES_NOT_EXIST_domain_resource_list(self, method, url, body, headers):
        body = self.fixtures.load("get_record_does_not_exist.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _domain_create(self, method, url, body, headers):
        body = self.fixtures.load("create_domain.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _VALIDATION_ERROR_domain_create(self, method, url, body, headers):
        body = self.fixtures.load("create_domain_validation_error.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _domain_update(self, method, url, body, headers):
        body = self.fixtures.load("update_domain.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _domain_resource_create(self, method, url, body, headers):
        body = self.fixtures.load("create_resource.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _domain_resource_update(self, method, url, body, headers):
        body = self.fixtures.load("update_resource.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _domain_delete(self, method, url, body, headers):
        body = self.fixtures.load("delete_domain.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _ZONE_DOES_NOT_EXIST_domain_delete(self, method, url, body, headers):
        body = self.fixtures.load("delete_domain_does_not_exist.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _domain_resource_delete(self, method, url, body, headers):
        body = self.fixtures.load("delete_resource.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _RECORD_DOES_NOT_EXIST_domain_resource_delete(self, method, url, body, headers):
        body = self.fixtures.load("delete_resource_does_not_exist.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import shlex
import base64
import datetime

from libcloud.utils.py3 import b, httplib
from libcloud.common.base import JsonResponse, ConnectionUserAndKey, KeyCertificateConnection
from libcloud.common.types import InvalidCredsError
from libcloud.container.base import Container, ContainerImage, ContainerDriver
from libcloud.container.types import ContainerState
from libcloud.container.providers import Provider

try:
    import simplejson as json
except Exception:
    import json


VALID_RESPONSE_CODES = [
    httplib.OK,
    httplib.ACCEPTED,
    httplib.CREATED,
    httplib.NO_CONTENT,
]


class DockerResponse(JsonResponse):
    valid_response_codes = [
        httplib.OK,
        httplib.ACCEPTED,
        httplib.CREATED,
        httplib.NO_CONTENT,
    ]

    def parse_body(self):
        if len(self.body) == 0 and not self.parse_zero_length_body:
            return self.body

        try:
            # error responses are tricky in Docker. Eg response could be
            # an error, but response status could still be 200
            content_type = self.headers.get("content-type", "application/json")
            if content_type == "application/json" or content_type == "":
                if (
                    self.headers.get("transfer-encoding") == "chunked"
                    and "fromImage" in self.request.url
                ):
                    body = [
                        json.loads(chunk)
                        for chunk in self.body.strip().replace("\r", "").split("\n")
                    ]
                else:
                    body = json.loads(self.body)
            else:
                body = self.body
        except ValueError:
            m = re.search('Error: (.+?)"', self.body)
            if m:
                error_msg = m.group(1)
                raise Exception(error_msg)
            else:
                raise Exception("ConnectionError: Failed to parse JSON response")
        return body

    def parse_error(self):
        if self.status == 401:
            raise InvalidCredsError("Invalid credentials")
        return self.body

    def success(self):
        return self.status in self.valid_response_codes


class DockerException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return "{} {}".format(self.code, self.message)

    def __repr__(self):
        return "DockerException {} {}".format(self.code, self.message)


class DockerConnection(ConnectionUserAndKey):
    responseCls = DockerResponse
    timeout = 60

    def add_default_headers(self, headers):
        """
        Add parameters that are necessary for every request
        If user and password are specified, include a base http auth
        header
        """
        headers["Content-Type"] = "application/json"
        if self.user_id and self.key:
            user_b64 = base64.b64encode(b("{}:{}".format(self.user_id, self.key)))
            headers["Authorization"] = "Basic %s" % (user_b64.decode("utf-8"))
        return headers


class DockertlsConnection(KeyCertificateConnection):
    responseCls = DockerResponse

    def __init__(
        self,
        key,
        secret,
        secure=True,
        host="localhost",
        port=4243,
        key_file="",
        cert_file="",
        **kwargs,
    ):
        super().__init__(
            key_file=key_file,
            cert_file=cert_file,
            secure=secure,
            host=host,
            port=port,
            url=None,
            proxy_url=None,
            timeout=None,
            backoff=None,
            retry_delay=None,
        )
        if key_file:
            keypath = os.path.expanduser(key_file)
            is_file_path = os.path.exists(keypath) and os.path.isfile(keypath)
            if not is_file_path:
                raise InvalidCredsError(
                    "You need an key PEM file to authenticate with "
                    "Docker tls. This can be found in the server."
                )
            self.key_file = key_file

            certpath = os.path.expanduser(cert_file)
            is_file_path = os.path.exists(certpath) and os.path.isfile(certpath)
            if not is_file_path:
                raise InvalidCredsError(
                    "You need an certificate PEM file to authenticate with "
                    "Docker tls. This can be found in the server."
                )
            self.cert_file = cert_file

    def add_default_headers(self, headers):
        headers["Content-Type"] = "application/json"
        return headers


class DockerContainerDriver(ContainerDriver):
    """
    Docker container driver class.

    >>> from libcloud.container.providers import get_driver
    >>> driver = get_driver('docker')
    >>> conn = driver(host='198.61.239.128', port=4243)
    >>> conn.list_containers()
    or connecting to http basic auth protected https host:
    >>> conn = driver('user', 'pass', host='https://198.61.239.128', port=443)

    connect with tls authentication, by providing a hostname, port, a private
    key file (.pem) and certificate (.pem) file
    >>> conn = driver(host='https://198.61.239.128',
    >>> port=4243, key_file='key.pem', cert_file='cert.pem')
    """

    type = Provider.DOCKER
    name = "Docker"
    website = "http://docker.io"
    connectionCls = DockerConnection
    supports_clusters = False
    version = "1.24"

    def __init__(
        self,
        key="",
        secret="",
        secure=False,
        host="localhost",
        port=4243,
        key_file=None,
        cert_file=None,
    ):
        """
        :param    key: API key or username to used (required)
        :type     key: ``str``

        :param    secret: Secret password to be used (required)
        :type     secret: ``str``

        :param    secure: Whether to use HTTPS or HTTP. Note: Some providers
                only support HTTPS, and it is on by default.
        :type     secure: ``bool``

        :param    host: Override hostname used for connections.
        :type     host: ``str``

        :param    port: Override port used for connections.
        :type     port: ``int``

        :param    key_file: Path to private key for TLS connection (optional)
        :type     key_file: ``str``

        :param    cert_file: Path to public key for TLS connection (optional)
        :type     cert_file: ``str``

        :return: ``None``
        """
        if key_file:
            self.connectionCls = DockertlsConnection
            self.key_file = key_file
            self.cert_file = cert_file
            secure = True

        if host.startswith("https://"):
            secure = True

        # strip the prefix
        prefixes = ["http://", "https://"]
        for prefix in prefixes:
            if host.startswith(prefix):
                host = host.strip(prefix)

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            key_file=key_file,
            cert_file=cert_file,
        )

        if key_file or cert_file:
            # docker tls authentication-
            # https://docs.docker.com/articles/https/
            # We pass two files, a key_file with the
            # private key and cert_file with the certificate
            # libcloud will handle them through LibcloudHTTPSConnection
            if not (key_file and cert_file):
                raise Exception(
                    "Needs both private key file and " "certificate file for tls authentication"
                )

        self.connection.secure = secure
        self.connection.host = host
        self.connection.port = port
        # set API version
        self.version = self._get_api_version()

    def _ex_connection_class_kwargs(self):
        kwargs = {}
        if hasattr(self, "key_file"):
            kwargs["key_file"] = self.key_file
        if hasattr(self, "cert_file"):
            kwargs["cert_file"] = self.cert_file
        return kwargs

    def install_image(self, path):
        """
        Install a container image from a remote path.

        :param path: Path to the container image
        :type  path: ``str``

        :rtype: :class:`libcloud.container.base.ContainerImage`
        """
        payload = {}
        data = json.dumps(payload)

        result = self.connection.request(
            "/v{}/images/create?fromImage={}".format(self.version, path),
            data=data,
            method="POST",
        )
        if "errorDetail" in result.body:
            raise DockerException(None, result.body)
        image_id = None

        # the response is slightly different if the image is already present
        # and it's not downloaded. both messages below indicate that the image
        # is available for use to the daemon
        if re.search(r"Downloaded newer image", result.body) or re.search(
            r'"Status: Image is up to date', result.body
        ):
            if re.search(r"sha256:(?P<id>[a-z0-9]{64})", result.body):
                image_id = re.findall(r"sha256:(?P<id>[a-z0-9]{64})", result.body)[-1]

        # if there is a failure message or if there is not an image id in the
        # response then throw an exception.
        if image_id is None:
            raise DockerException(None, "failed to install image")

        image = ContainerImage(
            id=image_id,
            name=path,
            path=path,
            version=None,
            driver=self.connection.driver,
            extra={},
        )
        return image

    def list_images(self):
        """
        List the installed container images

        :rtype: ``list`` of :class:`libcloud.container.base.ContainerImage`
        """
        result = self.connection.request("/v%s/images/json" % (self.version)).object
        images = []
        for image in result:
            try:
                name = image.get("RepoTags")[0]
            except Exception:
                name = image.get("Id")
            images.append(
                ContainerImage(
                    id=image.get("Id"),
                    name=name,
                    path=name,
                    version=None,
                    driver=self.connection.driver,
                    extra={
                        "created": image.get("Created"),
                        "size": image.get("Size"),
                        "virtual_size": image.get("VirtualSize"),
                    },
                )
            )

        return images

    def list_containers(self, image=None, all=True):
        """
        List the deployed container images

        :param image: Filter to containers with a certain image
        :type  image: :class:`libcloud.container.base.ContainerImage`

        :param all: Show all container (including stopped ones)
        :type  all: ``bool``

        :rtype: ``list`` of :class:`libcloud.container.base.Container`
        """
        if all:
            ex = "?all=1"
        else:
            ex = ""
        try:
            result = self.connection.request(
                "/v{}/containers/json{}".format(self.version, ex)
            ).object
        except Exception as exc:
            errno = getattr(exc, "errno", None)
            if errno == 111:
                raise DockerException(
                    errno,
                    "Make sure docker host is accessible" "and the API port is correct",
                )
            raise

        containers = [self._to_container(value) for value in result]
        return containers

    def deploy_container(
        self,
        name,
        image,
        parameters=None,
        start=True,
        command=None,
        hostname=None,
        user="",
        stdin_open=True,
        tty=True,
        mem_limit=0,
        ports=None,
        environment=None,
        dns=None,
        volumes=None,
        volumes_from=None,
        network_disabled=False,
        entrypoint=None,
        cpu_shares=None,
        working_dir="",
        domainname=None,
        memswap_limit=0,
        port_bindings=None,
        network_mode="bridge",
        labels=None,
    ):
        """
        Deploy an installed container image

        For details on the additional parameters see : http://bit.ly/1PjMVKV

        :param name: The name of the new container
        :type  name: ``str``

        :param image: The container image to deploy
        :type  image: :class:`libcloud.container.base.ContainerImage`

        :param parameters: Container Image parameters
        :type  parameters: ``str``

        :param start: Start the container on deployment
        :type  start: ``bool``

        :rtype: :class:`Container`
        """
        command = shlex.split(str(command))
        if port_bindings is None:
            port_bindings = {}
        params = {"name": name}

        payload = {
            "Hostname": hostname,
            "Domainname": domainname,
            "ExposedPorts": ports,
            "User": user,
            "Tty": tty,
            "OpenStdin": stdin_open,
            "StdinOnce": False,
            "Memory": mem_limit,
            "AttachStdin": True,
            "AttachStdout": True,
            "AttachStderr": True,
            "Env": environment,
            "Cmd": command,
            "Dns": dns,
            "Image": image.name,
            "Volumes": volumes,
            "VolumesFrom": volumes_from,
            "NetworkDisabled": network_disabled,
            "Entrypoint": entrypoint,
            "CpuShares": cpu_shares,
            "WorkingDir": working_dir,
            "MemorySwap": memswap_limit,
            "PublishAllPorts": True,
            "PortBindings": port_bindings,
            "NetworkMode": network_mode,
            "Labels": labels,
        }

        data = json.dumps(payload)
        try:
            result = self.connection.request(
                "/v%s/containers/create" % (self.version),
                data=data,
                params=params,
                method="POST",
            )
        except Exception as e:
            message = e.message or str(e)  # pylint: disable=no-member
            if message.startswith("No such image:"):
                raise DockerException(None, "No such image: %s" % image.name)
            else:
                raise DockerException(None, e)

        id_ = result.object["Id"]

        payload = {
            "Binds": [],
            "PublishAllPorts": True,
            "PortBindings": port_bindings,
        }

        data = json.dumps(payload)
        if start:
            if float(self._get_api_version()) > 1.22:
                result = self.connection.request(
                    "/v{}/containers/{}/start".format(self.version, id_), method="POST"
                )
            else:
                result = self.connection.request(
                    "/v{}/containers/{}/start".format(self.version, id_),
                    data=data,
                    method="POST",
                )

        return self.get_container(id_)

    def get_container(self, id):
        """
        Get a container by ID

        :param id: The ID of the container to get
        :type  id: ``str``

        :rtype: :class:`libcloud.container.base.Container`
        """
        result = self.connection.request("/v{}/containers/{}/json".format(self.version, id)).object

        return self._to_container(result)

    def start_container(self, container):
        """
        Start a container

        :param container: The container to be started
        :type  container: :class:`libcloud.container.base.Container`

        :return: The container refreshed with current data
        :rtype: :class:`libcloud.container.base.Container`
        """
        if float(self._get_api_version()) > 1.22:
            result = self.connection.request(
                "/v{}/containers/{}/start".format(self.version, container.id), method="POST"
            )
        else:
            payload = {
                "Binds": [],
                "PublishAllPorts": True,
            }
            data = json.dumps(payload)
            result = self.connection.request(
                "/v{}/containers/{}/start".format(self.version, container.id),
                method="POST",
                data=data,
            )

        if result.status in VALID_RESPONSE_CODES:
            return self.get_container(container.id)
        else:
            raise DockerException(result.status, "failed to start container")

    def stop_container(self, container):
        """
        Stop a container

        :param container: The container to be stopped
        :type  container: :class:`libcloud.container.base.Container`

        :return: The container refreshed with current data
        :rtype: :class:`libcloud.container.base.Container`
        """
        result = self.connection.request(
            "/v{}/containers/{}/stop".format(self.version, container.id), method="POST"
        )
        if result.status in VALID_RESPONSE_CODES:
            return self.get_container(container.id)
        else:
            raise DockerException(result.status, "failed to stop container")

    def restart_container(self, container):
        """
        Restart a container

        :param container: The container to be stopped
        :type  container: :class:`libcloud.container.base.Container`

        :return: The container refreshed with current data
        :rtype: :class:`libcloud.container.base.Container`
        """
        data = json.dumps({"t": 10})
        # number of seconds to wait before killing the container
        result = self.connection.request(
            "/v{}/containers/{}/restart".format(self.version, container.id),
            data=data,
            method="POST",
        )
        if result.status in VALID_RESPONSE_CODES:
            return self.get_container(container.id)
        else:
            raise DockerException(result.status, "failed to restart container")

    def destroy_container(self, container):
        """
        Remove a container

        :param container: The container to be destroyed
        :type  container: :class:`libcloud.container.base.Container`

        :return: True if the destroy was successful, False otherwise.
        :rtype: ``bool``
        """
        result = self.connection.request(
            "/v{}/containers/{}".format(self.version, container.id), method="DELETE"
        )
        return result.status in VALID_RESPONSE_CODES

    def ex_list_processes(self, container):
        """
        List processes running inside a container

        :param container: The container to list processes for.
        :type  container: :class:`libcloud.container.base.Container`

        :rtype: ``str``
        """
        result = self.connection.request(
            "/v{}/containers/{}/top".format(self.version, container.id)
        ).object

        return result

    def ex_rename_container(self, container, name):
        """
        Rename a container

        :param container: The container to be renamed
        :type  container: :class:`libcloud.container.base.Container`

        :param name: The new name
        :type  name: ``str``

        :rtype: :class:`libcloud.container.base.Container`
        """
        result = self.connection.request(
            "/v{}/containers/{}/rename?name={}".format(self.version, container.id, name),
            method="POST",
        )
        if result.status in VALID_RESPONSE_CODES:
            return self.get_container(container.id)

    def ex_get_logs(self, container, stream=False):
        """
        Get container logs

        If stream == True, logs will be yielded as a stream
        From Api Version 1.11 and above we need a GET request to get the logs
        Logs are in different format of those of Version 1.10 and below

        :param container: The container to list logs for
        :type  container: :class:`libcloud.container.base.Container`

        :param stream: Stream the output
        :type  stream: ``bool``

        :rtype: ``bool``
        """
        payload = {}
        data = json.dumps(payload)

        if float(self._get_api_version()) > 1.10:
            result = self.connection.request(
                "/v%s/containers/%s/logs?follow=%s&stdout=1&stderr=1"
                % (self.version, container.id, str(stream))
            ).object
            logs = result
        else:
            result = self.connection.request(
                "/v%s/containers/%s/attach?logs=1&stream=%s&stdout=1&stderr=1"
                % (self.version, container.id, str(stream)),
                method="POST",
                data=data,
            )
            logs = result.body

        return logs

    def ex_search_images(self, term):
        """Search for an image on Docker.io.
        Returns a list of ContainerImage objects

        >>> images = conn.ex_search_images(term='mistio')
        >>> images
        [<ContainerImage: id=rolikeusch/docker-mistio...>,
         <ContainerImage: id=mist/mistio, name=mist/mistio,
             driver=Docker  ...>]

         :param term: The search term
         :type  term: ``str``

         :rtype: ``list`` of :class:`libcloud.container.base.ContainerImage`
        """

        term = term.replace(" ", "+")
        result = self.connection.request(
            "/v{}/images/search?term={}".format(self.version, term)
        ).object
        images = []
        for image in result:
            name = image.get("name")
            images.append(
                ContainerImage(
                    id=name,
                    path=name,
                    version=None,
                    name=name,
                    driver=self.connection.driver,
                    extra={
                        "description": image.get("description"),
                        "is_official": image.get("is_official"),
                        "is_trusted": image.get("is_trusted"),
                        "star_count": image.get("star_count"),
                    },
                )
            )

        return images

    def ex_delete_image(self, image):
        """
        Remove image from the filesystem

        :param  image: The image to remove
        :type   image: :class:`libcloud.container.base.ContainerImage`

        :rtype: ``bool``
        """
        result = self.connection.request(
            "/v{}/images/{}".format(self.version, image.name), method="DELETE"
        )
        return result.status in VALID_RESPONSE_CODES

    def _to_container(self, data):
        """
        Convert container in Container instances
        """
        try:
            name = data.get("Name").strip("/")
        except Exception:
            try:
                name = data.get("Names")[0].strip("/")
            except Exception:
                name = data.get("Id")
        state = data.get("State")
        if isinstance(state, dict):
            status = data.get("Status", state.get("Status") if state is not None else None)
        else:
            status = data.get("Status")
        if "Exited" in status:
            state = ContainerState.STOPPED
        elif status.startswith("Up "):
            state = ContainerState.RUNNING
        elif "running" in status:
            state = ContainerState.RUNNING
        else:
            state = ContainerState.STOPPED
        image = data.get("Image")
        ports = data.get("Ports", [])
        created = data.get("Created")
        if isinstance(created, float):
            created = ts_to_str(created)
        extra = {
            "id": data.get("Id"),
            "status": data.get("Status"),
            "created": created,
            "image": image,
            "ports": ports,
            "command": data.get("Command"),
            "sizerw": data.get("SizeRw"),
            "sizerootfs": data.get("SizeRootFs"),
        }
        ips = []
        if ports is not None:
            for port in ports:
                if port.get("IP") is not None:
                    ips.append(port.get("IP"))
        return Container(
            id=data["Id"],
            name=name,
            image=ContainerImage(
                id=data.get("ImageID", None),
                path=image,
                name=image,
                version=None,
                driver=self.connection.driver,
            ),
            ip_addresses=ips,
            state=state,
            driver=self.connection.driver,
            extra=extra,
        )

    def _get_api_version(self):
        """
        Get the docker API version information
        """
        result = self.connection.request("/version").object
        result = result or {}
        api_version = result.get("ApiVersion")

        return api_version


def ts_to_str(timestamp):
    """
    Return a timestamp as a nicely formatted datetime string.
    """
    date = datetime.datetime.fromtimestamp(timestamp)
    date_string = date.strftime("%d/%m/%Y %H:%M %Z")
    return date_string



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
import sys
import json
import base64

from libcloud.test import MockHttp, LibcloudTestCase, unittest
from libcloud.compute import providers
from libcloud.utils.py3 import httplib, ensure_string
from libcloud.common.types import InvalidCredsError
from libcloud.compute.base import Node, NodeSize, NodeImage, NodeLocation, NodeAuthSSHKey
from libcloud.test.secrets import UPCLOUD_PARAMS
from libcloud.compute.types import Provider, NodeState
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.upcloud import UpcloudDriver, UpcloudResponse


class UpcloudPersistResponse(UpcloudResponse):
    def parse_body(self):
        import os

        path = os.path.abspath(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                os.path.pardir,
                "compute",
                "fixtures",
                "upcloud",
            )
        )
        filename = "api" + self.request.path_url.replace("/", "_").replace(".", "_") + ".json"
        filename = os.path.join(path, filename)
        if not os.path.exists(filename):
            with open(filename, "w+") as f:
                f.write(self.body)
        return super().parse_body()


class UpcloudAuthenticationTests(LibcloudTestCase):
    def setUp(self):
        UpcloudDriver.connectionCls.conn_class = UpcloudMockHttp
        self.driver = UpcloudDriver("nosuchuser", "nopwd")

    def test_authentication_fails(self):
        with self.assertRaises(InvalidCredsError):
            self.driver.list_locations()


class UpcloudDriverTests(LibcloudTestCase):
    def setUp(self):
        UpcloudDriver.connectionCls.conn_class = UpcloudMockHttp
        # UpcloudDriver.connectionCls.responseCls = UpcloudPersistResponse
        self.driver = UpcloudDriver(*UPCLOUD_PARAMS)

    def test_creating_driver(self):
        cls = providers.get_driver(Provider.UPCLOUD)
        self.assertIs(cls, UpcloudDriver)

    def test_features(self):
        features = self.driver.features["create_node"]
        self.assertIn("ssh_key", features)
        self.assertIn("generates_password", features)

    def test_list_locations(self):
        locations = self.driver.list_locations()
        self.assertTrue(len(locations) >= 1)
        expected_node_location = NodeLocation(
            id="fi-hel1", name="Helsinki #1", country="FI", driver=self.driver
        )
        self.assert_object(expected_node_location, objects=locations)

    def test_list_sizes(self):
        location = NodeLocation(id="fi-hel1", name="Helsinki #1", country="FI", driver=self.driver)
        sizes = self.driver.list_sizes(location)
        self.assertTrue(len(sizes) >= 1)
        expected_node_size = NodeSize(
            id="1xCPU-1GB",
            name="1xCPU-1GB",
            ram=1024,
            disk=30,
            bandwidth=2048,
            price=2.232,
            driver=self.driver,
            extra={"core_number": 1, "storage_tier": "maxiops"},
        )
        self.assert_object(expected_node_size, objects=sizes)

    def test_list_images(self):
        images = self.driver.list_images()
        self.assertTrue(len(images) >= 1)
        expected_node_image = NodeImage(
            id="01000000-0000-4000-8000-000010010101",
            name="Windows Server 2003 R2 Standard (CD 1)",
            driver=self.driver,
            extra={
                "access": "public",
                "license": 0,
                "size": 1,
                "state": "online",
                "type": "cdrom",
            },
        )
        self.assert_object(expected_node_image, objects=images)

    def test_create_node_from_template(self):
        image = NodeImage(
            id="01000000-0000-4000-8000-000030060200",
            name="Ubuntu Server 16.04 LTS (Xenial Xerus)",
            extra={"type": "template"},
            driver=self.driver,
        )
        location = NodeLocation(id="fi-hel1", name="Helsinki #1", country="FI", driver=self.driver)
        size = NodeSize(
            id="1xCPU-1GB",
            name="1xCPU-1GB",
            ram=1024,
            disk=30,
            bandwidth=2048,
            extra={"storage_tier": "maxiops"},
            price=None,
            driver=self.driver,
        )
        node = self.driver.create_node(
            name="test_server",
            size=size,
            image=image,
            location=location,
            ex_hostname="myhost.somewhere",
        )

        self.assertTrue(
            re.match(
                "^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$",
                node.id,
            )
        )
        self.assertEqual(node.name, "test_server")
        self.assertEqual(node.state, NodeState.STARTING)
        self.assertTrue(len(node.public_ips) > 0)
        self.assertTrue(len(node.private_ips) > 0)
        self.assertEqual(node.driver, self.driver)
        self.assertTrue(len(node.extra["password"]) > 0)
        self.assertTrue(len(node.extra["vnc_password"]) > 0)

    def test_create_node_with_ssh_keys(self):
        image = NodeImage(
            id="01000000-0000-4000-8000-000030060200",
            name="Ubuntu Server 16.04 LTS (Xenial Xerus)",
            extra={"type": "template"},
            driver=self.driver,
        )
        location = NodeLocation(id="fi-hel1", name="Helsinki #1", country="FI", driver=self.driver)
        size = NodeSize(
            id="1xCPU-1GB",
            name="1xCPU-1GB",
            ram=1024,
            disk=30,
            bandwidth=2048,
            extra={"storage_tier": "maxiops"},
            price=None,
            driver=self.driver,
        )

        auth = NodeAuthSSHKey("publikey")
        node = self.driver.create_node(
            name="test_server", size=size, image=image, location=location, auth=auth
        )
        self.assertTrue(
            re.match(
                "^[a-z0-9]{8}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{4}-[a-z0-9]{12}$",
                node.id,
            )
        )
        self.assertEqual(node.name, "test_server")
        self.assertEqual(node.state, NodeState.STARTING)
        self.assertTrue(len(node.public_ips) > 0)
        self.assertTrue(len(node.private_ips) > 0)
        self.assertEqual(node.driver, self.driver)

    def test_list_nodes(self):
        nodes = self.driver.list_nodes()

        self.assertTrue(len(nodes) >= 1)
        node = nodes[0]
        self.assertEqual(node.name, "test_server")
        self.assertEqual(node.state, NodeState.RUNNING)
        self.assertTrue(len(node.public_ips) > 0)
        self.assertTrue(len(node.private_ips) > 0)
        self.assertEqual(node.driver, self.driver)

    def test_reboot_node(self):
        nodes = self.driver.list_nodes()
        success = self.driver.reboot_node(nodes[0])
        self.assertTrue(success)

    def test_destroy_node(self):
        if UpcloudDriver.connectionCls.conn_class == UpcloudMockHttp:
            nodes = [
                Node(
                    id="00893c98_5d5a_4363_b177_88df518a2b60",
                    name="",
                    state="",
                    public_ips=[],
                    private_ips=[],
                    driver=self.driver,
                )
            ]
        else:
            nodes = self.driver.list_nodes()
        success = self.driver.destroy_node(nodes[0])
        self.assertTrue(success)

    def assert_object(self, expected_object, objects):
        same_data = any([self.objects_equals(expected_object, obj) for obj in objects])
        self.assertTrue(same_data, "Objects does not match")

    def objects_equals(self, expected_obj, obj):
        for name in vars(expected_obj):
            expected_data = getattr(expected_obj, name)
            actual_data = getattr(obj, name)
            same_data = self.data_equals(expected_data, actual_data)
            if not same_data:
                break
        return same_data

    def data_equals(self, expected_data, actual_data):
        if isinstance(expected_data, dict):
            return self.dicts_equals(expected_data, actual_data)
        else:
            return expected_data == actual_data

    def dicts_equals(self, d1, d2):
        dict_keys_same = set(d1.keys()) == set(d2.keys())
        if not dict_keys_same:
            return False

        for key in d1.keys():
            if d1[key] != d2[key]:
                return False

        return True


class UpcloudMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("upcloud")

    def _1_2_zone(self, method, url, body, headers):
        auth = headers["Authorization"].split(" ")[1]
        username, password = ensure_string(base64.b64decode(auth)).split(":")
        if username == "nosuchuser" and password == "nopwd":
            body = self.fixtures.load("api_1_2_zone_failed_auth.json")
            status_code = httplib.UNAUTHORIZED
        else:
            body = self.fixtures.load("api_1_2_zone.json")
            status_code = httplib.OK
        return (status_code, body, {}, httplib.responses[httplib.OK])

    def _1_2_plan(self, method, url, body, headers):
        body = self.fixtures.load("api_1_2_plan.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _1_2_storage_cdrom(self, method, url, body, headers):
        body = self.fixtures.load("api_1_2_storage_cdrom.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _1_2_storage_template(self, method, url, body, headers):
        body = self.fixtures.load("api_1_2_storage_template.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _1_2_price(self, method, url, body, headers):
        body = self.fixtures.load("api_1_2_price.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _1_2_server(self, method, url, body, headers):
        if method == "POST":
            dbody = json.loads(body)
            storages = dbody["server"]["storage_devices"]["storage_device"]
            if any(["type" in storage and storage["type"] == "cdrom" for storage in storages]):
                body = self.fixtures.load("api_1_2_server_from_cdrom.json")
            else:
                body = self.fixtures.load("api_1_2_server_from_template.json")
        else:
            body = self.fixtures.load("api_1_2_server.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _1_2_server_00f8c525_7e62_4108_8115_3958df5b43dc(self, method, url, body, headers):
        body = self.fixtures.load("api_1_2_server_00f8c525-7e62-4108-8115-3958df5b43dc.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _1_2_server_00f8c525_7e62_4108_8115_3958df5b43dc_restart(self, method, url, body, headers):
        body = self.fixtures.load(
            "api_1_2_server_00f8c525-7e62-4108-8115-3958df5b43dc_restart.json"
        )
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _1_2_server_00893c98_5d5a_4363_b177_88df518a2b60(self, method, url, body, headers):
        body = self.fixtures.load("api_1_2_server_00893c98-5d5a-4363-b177-88df518a2b60.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
from typing import Any, Dict, List, Type, Union, Iterator, Optional

from libcloud import __version__
from libcloud.dns.types import RecordType
from libcloud.common.base import BaseDriver, Connection, ConnectionUserAndKey

__all__ = ["Zone", "Record", "DNSDriver"]


class Zone:
    """
    DNS zone.
    """

    def __init__(
        self,
        id,  # type: str
        domain,  # type: str
        type,  # type: str
        ttl,  # type: int
        driver,  # type: DNSDriver
        extra=None,  # type: Optional[dict]
    ):
        """
        :param id: Zone id.
        :type id: ``str``

        :param domain: The name of the domain.
        :type domain: ``str``

        :param type: Zone type (master, slave).
        :type type: ``str``

        :param ttl: Default TTL for records in this zone (in seconds).
        :type ttl: ``int``

        :param driver: DNSDriver instance.
        :type driver: :class:`DNSDriver`

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``
        """
        self.id = str(id) if id else None
        self.domain = domain
        self.type = type
        self.ttl = ttl or None
        self.driver = driver
        self.extra = extra or {}

    def list_records(self):
        # type: () -> List[Record]
        return self.driver.list_records(zone=self)

    def create_record(self, name, type, data, extra=None):
        # type: (str, RecordType, str, Optional[dict]) -> Record
        return self.driver.create_record(name=name, zone=self, type=type, data=data, extra=extra)

    def update(
        self,
        domain=None,  # type: Optional[str]
        type=None,  # type: Optional[str]
        ttl=None,  # type: Optional[int]
        extra=None,  # type: Optional[dict]
    ):
        # type: (...) -> Zone
        return self.driver.update_zone(zone=self, domain=domain, type=type, ttl=ttl, extra=extra)

    def delete(self):
        # type: () -> bool
        return self.driver.delete_zone(zone=self)

    def export_to_bind_format(self):
        # type: () -> str
        return self.driver.export_zone_to_bind_format(zone=self)

    def export_to_bind_zone_file(self, file_path):
        # type: (str) -> None
        self.driver.export_zone_to_bind_zone_file(zone=self, file_path=file_path)

    def __repr__(self):
        # type: () -> str
        return "<Zone: domain={}, ttl={}, provider={} ...>".format(
            self.domain,
            self.ttl,
            self.driver.name,
        )


class Record:
    """
    Zone record / resource.
    """

    def __init__(
        self,
        id,  # type: str
        name,  # type: str
        type,  # type: RecordType
        data,  # type: str
        zone,  # type: Zone
        driver,  # type: DNSDriver
        ttl=None,  # type: Optional[int]
        extra=None,  # type: Optional[dict]
    ):
        """
        :param id: Record id
        :type id: ``str``

        :param name: Hostname or FQDN.
        :type name: ``str``

        :param type: DNS record type (A, AAAA, ...).
        :type type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type data: ``str``

        :param zone: Zone instance.
        :type zone: :class:`Zone`

        :param driver: DNSDriver instance.
        :type driver: :class:`DNSDriver`

        :param ttl: Record TTL.
        :type ttl: ``int``

        :param extra: (optional) Extra attributes (driver specific).
        :type extra: ``dict``
        """
        self.id = str(id) if id else None
        self.name = name
        self.type = type
        self.data = data
        self.zone = zone
        self.driver = driver
        self.ttl = ttl
        self.extra = extra or {}

    def update(
        self,
        name=None,  # type: Optional[str]
        type=None,  # type: Optional[RecordType]
        data=None,  # type: Optional[str]
        extra=None,  # type: Optional[dict]
    ):
        # type: (...) -> Record
        return self.driver.update_record(record=self, name=name, type=type, data=data, extra=extra)

    def delete(self):
        # type: () -> bool
        return self.driver.delete_record(record=self)

    def _get_numeric_id(self):
        # type: () -> Union[int, str]
        """
        Return numeric ID for the provided record if the ID is a digit.

        This method is used for sorting the values when exporting Zone to a
        BIND format.
        """
        record_id = self.id

        if record_id is None:
            return ""

        if record_id.isdigit():
            record_id_int = int(record_id)
            return record_id_int

        return record_id

    def __repr__(self):
        # type: () -> str
        zone = self.zone.domain if self.zone.domain else self.zone.id
        return "<Record: zone=%s, name=%s, type=%s, data=%s, provider=%s, " "ttl=%s ...>" % (
            zone,
            self.name,
            self.type,
            self.data,
            self.driver.name,
            self.ttl,
        )


class DNSDriver(BaseDriver):
    """
    A base DNSDriver class to derive from

    This class is always subclassed by a specific driver.
    """

    connectionCls = ConnectionUserAndKey  # type: Type[Connection]
    name = None  # type: str
    website = None  # type: str

    # Map libcloud record type enum to provider record type name
    RECORD_TYPE_MAP = {}  # type: Dict[RecordType, str]

    def __init__(
        self,
        key,  # type: str
        secret=None,  # type: Optional[str]
        secure=True,  # type: bool
        host=None,  # type: Optional[str]
        port=None,  # type: Optional[int]
        **kwargs,  # type: Optional[Any]
    ):
        # type: (...) -> None
        """
        :param    key: API key or username to used (required)
        :type     key: ``str``

        :param    secret: Secret password to be used (required)
        :type     secret: ``str``

        :param    secure: Whether to use HTTPS or HTTP. Note: Some providers
                only support HTTPS, and it is on by default.
        :type     secure: ``bool``

        :param    host: Override hostname used for connections.
        :type     host: ``str``

        :param    port: Override port used for connections.
        :type     port: ``int``

        :return: ``None``
        """
        super().__init__(key=key, secret=secret, secure=secure, host=host, port=port, **kwargs)

    def list_record_types(self):
        # type: () -> List[RecordType]
        """
        Return a list of RecordType objects supported by the provider.

        :return: ``list`` of :class:`RecordType`
        """
        return list(self.RECORD_TYPE_MAP.keys())

    def iterate_zones(self):
        # type: () -> Iterator[Zone]
        """
        Return a generator to iterate over available zones.

        :rtype: ``generator`` of :class:`Zone`
        """
        raise NotImplementedError("iterate_zones not implemented for this driver")

    def list_zones(self):
        # type: () -> List[Zone]
        """
        Return a list of zones.

        :return: ``list`` of :class:`Zone`
        """
        return list(self.iterate_zones())

    def iterate_records(self, zone):
        # type: (Zone) -> Iterator[Record]
        """
        Return a generator to iterate over records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :rtype: ``generator`` of :class:`Record`
        """
        raise NotImplementedError("iterate_records not implemented for this driver")

    def list_records(self, zone):
        # type: (Zone) -> List[Record]
        """
        Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        return list(self.iterate_records(zone))

    def get_zone(self, zone_id):
        # type: (str) -> Zone
        """
        Return a Zone instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        raise NotImplementedError("get_zone not implemented for this driver")

    def get_record(self, zone_id, record_id):
        # type: (str, str) -> Record
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        raise NotImplementedError("get_record not implemented for this driver")

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        # type: (str, str, Optional[int], Optional[dict]) -> Zone
        """
        Create a new zone.

        :param domain: Zone domain name (e.g. example.com)
        :type domain: ``str``

        :param type: Zone type (master / slave).
        :type  type: ``str``

        :param ttl: TTL for new records. (optional)
        :type  ttl: ``int``

        :param extra: Extra attributes (driver specific). (optional)
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """
        raise NotImplementedError("create_zone not implemented for this driver")

    def update_zone(
        self,
        zone,  # type: Zone
        domain,  # type: Optional[str]
        type="master",  # type: Optional[str]
        ttl=None,  # type: Optional[int]
        extra=None,  # type: Optional[dict]
    ):
        # type: (...) -> Zone
        """
        Update an existing zone.

        :param zone: Zone to update.
        :type  zone: :class:`Zone`

        :param domain: Zone domain name (e.g. example.com)
        :type  domain: ``str``

        :param type: Zone type (master / slave).
        :type  type: ``str``

        :param ttl: TTL for new records. (optional)
        :type  ttl: ``int``

        :param extra: Extra attributes (driver specific). (optional)
        :type  extra: ``dict``

        :rtype: :class:`Zone`
        """
        raise NotImplementedError("update_zone not implemented for this driver")

    def create_record(self, name, zone, type, data, extra=None):
        # type: (str, Zone, RecordType, str, Optional[dict]) -> Record
        """
        Create a new record.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Extra attributes (driver specific). (optional)
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        raise NotImplementedError("create_record not implemented for this driver")

    def update_record(
        self,
        record,  # type: Record
        name,  # type: Optional[str]
        type,  # type: Optional[RecordType]
        data,  # type: Optional[str]
        extra=None,  # type: Optional[dict]
    ):
        """
        Update an existing record.

        :param record: Record to update.
        :type  record: :class:`Record`

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: (optional) Extra attributes (driver specific).
        :type  extra: ``dict``

        :rtype: :class:`Record`
        """
        raise NotImplementedError("update_record not implemented for this driver")

    def delete_zone(self, zone):
        # type: (Zone) -> bool
        """
        Delete a zone.

        Note: This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        raise NotImplementedError("delete_zone not implemented for this driver")

    def delete_record(self, record):
        # type: (Record) -> bool
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        raise NotImplementedError("delete_record not implemented for this driver")

    def export_zone_to_bind_format(self, zone):
        # type: (Zone) -> str
        """
        Export Zone object to the BIND compatible format.

        :param zone: Zone to export.
        :type  zone: :class:`Zone`

        :return: Zone data in BIND compatible format.
        :rtype: ``str``
        """
        if zone.type != "master":
            raise ValueError("You can only generate BIND out for master zones")

        lines = []

        # For consistent output, records are sorted based on the id
        records = zone.list_records()
        records = sorted(records, key=Record._get_numeric_id)

        date = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        values = {"version": __version__, "date": date}

        lines.append("; Generated by Libcloud v%(version)s on %(date)s UTC" % values)
        lines.append("$ORIGIN {domain}.".format(domain=zone.domain))
        lines.append("$TTL {domain_ttl}\n".format(domain_ttl=zone.ttl))

        for record in records:
            line = self._get_bind_record_line(record=record)
            lines.append(line)

        output = "\n".join(lines)
        return output

    def export_zone_to_bind_zone_file(self, zone, file_path):
        # type: (Zone, str) -> None
        """
        Export Zone object to the BIND compatible format and write result to a
        file.

        :param zone: Zone to export.
        :type  zone: :class:`Zone`

        :param file_path: File path where the output will be saved.
        :type  file_path: ``str``
        """
        result = self.export_zone_to_bind_format(zone=zone)

        with open(file_path, "w") as fp:
            fp.write(result)

    def _get_bind_record_line(self, record):
        # type: (Record) -> str
        """
        Generate BIND record line for the provided record.

        :param record: Record to generate the line for.
        :type  record: :class:`Record`

        :return: Bind compatible record line.
        :rtype: ``str``
        """
        parts = []  # type: List[Any]

        if record.name:
            name = "{name}.{domain}".format(
                name=record.name,
                domain=record.zone.domain,
            )
        else:
            name = record.zone.domain

        name += "."

        ttl = record.extra["ttl"] if "ttl" in record.extra else record.zone.ttl
        ttl = str(ttl)
        data = record.data

        if record.type in [
            RecordType.CNAME,
            RecordType.DNAME,
            RecordType.MX,
            RecordType.PTR,
            RecordType.SRV,
        ]:
            # Make sure trailing dot is present
            if data[len(data) - 1] != ".":
                data += "."

        if record.type in [RecordType.TXT, RecordType.SPF] and " " in data:
            # Escape the quotes
            data = data.replace('"', '\\"')

            # Quote the string
            data = '"%s"' % (data)

        if record.type in [RecordType.MX, RecordType.SRV]:
            priority = str(record.extra["priority"])
            parts = [name, ttl, "IN", str(record.type), priority, data]
        else:
            parts = [name, ttl, "IN", str(record.type), data]

        line = "\t".join(parts)
        return line

    def _string_to_record_type(self, string):
        # type: (str) -> RecordType
        """
        Return a string representation of a DNS record type to a
        libcloud RecordType ENUM.

        :rtype: ``str``
        """
        string = string.upper()
        record_type = getattr(RecordType, string)
        return record_type



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.common.types import LazyList


class TestLazyList(unittest.TestCase):
    def setUp(self):
        super().setUp
        self._get_more_counter = 0

    def tearDown(self):
        super().tearDown

    def test_init(self):
        data = [1, 2, 3, 4, 5]
        ll = LazyList(get_more=self._get_more_exhausted)
        ll_list = list(ll)
        self.assertEqual(ll_list, data)

    def test_iterator(self):
        data = [1, 2, 3, 4, 5]
        ll = LazyList(get_more=self._get_more_exhausted)
        for i, d in enumerate(ll):
            self.assertEqual(d, data[i])

    def test_empty_list(self):
        ll = LazyList(get_more=self._get_more_empty)

        self.assertEqual(list(ll), [])
        self.assertEqual(len(ll), 0)
        self.assertTrue(10 not in ll)

    def test_iterator_not_exhausted(self):
        data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        ll = LazyList(get_more=self._get_more_not_exhausted)
        number_of_iterations = 0
        for i, d in enumerate(ll):
            self.assertEqual(d, data[i])
            number_of_iterations += 1
        self.assertEqual(number_of_iterations, 10)

    def test_len(self):
        ll = LazyList(get_more=self._get_more_not_exhausted)
        ll = LazyList(get_more=self._get_more_not_exhausted)

        self.assertEqual(len(ll), 10)

    def test_contains(self):
        ll = LazyList(get_more=self._get_more_not_exhausted)

        self.assertTrue(40 not in ll)
        self.assertTrue(1 in ll)
        self.assertTrue(5 in ll)
        self.assertTrue(10 in ll)

    def test_indexing(self):
        ll = LazyList(get_more=self._get_more_not_exhausted)

        self.assertEqual(ll[0], 1)
        self.assertEqual(ll[9], 10)
        self.assertEqual(ll[-1], 10)

        try:
            ll[11]
        except IndexError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_repr(self):
        ll1 = LazyList(get_more=self._get_more_empty)
        ll2 = LazyList(get_more=self._get_more_exhausted)
        ll3 = LazyList(get_more=self._get_more_not_exhausted)

        self.assertEqual(repr(ll1), "[]")
        self.assertEqual(repr(ll2), "[1, 2, 3, 4, 5]")
        self.assertEqual(repr(ll3), "[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]")

    def _get_more_empty(self, last_key, value_dict):
        return [], None, True

    def _get_more_exhausted(self, last_key, value_dict):
        data = [1, 2, 3, 4, 5]
        return data, 5, True

    def _get_more_not_exhausted(self, last_key, value_dict):
        self._get_more_counter += 1
        if not last_key:
            data, last_key, exhausted = [1, 2, 3, 4, 5], 5, False
        else:
            data, last_key, exhausted = [6, 7, 8, 9, 10], 10, True

        return data, last_key, exhausted


if __name__ == "__main__":
    sys.exit(unittest.main())
